@@ -5,7 +5,7 @@ bs_dir="$(dirname $(readlink -f $0))"
 touch "${bs_dir}"/ltmain.sh
 autoreconf -fi "${bs_dir}"
 
-if test -z "$NOCONFIGURE" ; then
+if test -n "$1" && test -z "$NOCONFIGURE" ; then
 	echo 'Configuring...'
 	"$bs_dir"/configure "$@"
 fi
