@@ -1,5 +1,5 @@
 /*
- * Copyright 2013 Con Kolivas <kernel@kolivas.org>
+ * Copyright 2013-2014 Con Kolivas <kernel@kolivas.org>
  * Copyright 2012-2014 Xiangfu <xiangfu@openmobilefree.com>
  * Copyright 2012 Luke Dashjr
  * Copyright 2012 Andrew Smith
@@ -17,7 +17,6 @@
 #include <stdio.h>
 #include <sys/time.h>
 #include <sys/types.h>
-#include <sys/select.h>
 #include <dirent.h>
 #include <unistd.h>
 #ifndef WIN32
@@ -37,38 +36,60 @@
 #include "fpgautils.h"
 #include "driver-avalon2.h"
 #include "crc.h"
-#include "hexdump.c"
+#include "sha2.h"
 
 #define ASSERT1(condition) __maybe_unused static char sizeof_uint32_t_must_be_4[(condition)?1:-1]
 ASSERT1(sizeof(uint32_t) == 4);
 
 #define get_fan_pwm(v)	(AVA2_PWM_MAX - (v) * AVA2_PWM_MAX / 100)
 
-int opt_avalon2_freq_min = AVA2_DEFAULT_FREQUENCY;
-int opt_avalon2_freq_max = AVA2_DEFAULT_FREQUENCY_MAX;
+int opt_avalon2_freq_min;
+int opt_avalon2_freq_max;
 
-int opt_avalon2_fan_min = AVA2_DEFAULT_FAN_PWM;
+int opt_avalon2_fan_min = AVA2_DEFAULT_FAN_MIN;
 int opt_avalon2_fan_max = AVA2_DEFAULT_FAN_MAX;
+static int avalon2_fan_min = get_fan_pwm(AVA2_DEFAULT_FAN_MIN);
+static int avalon2_fan_max = get_fan_pwm(AVA2_DEFAULT_FAN_MAX);
 
-int opt_avalon2_voltage_min = AVA2_DEFAULT_VOLTAGE;
-int opt_avalon2_voltage_max = AVA2_DEFAULT_VOLTAGE_MAX;
+int opt_avalon2_voltage_min;
+int opt_avalon2_voltage_max;
 
 int opt_avalon2_overheat = AVALON2_TEMP_OVERHEAT;
+int opt_avalon2_polling_delay = AVALON2_DEFAULT_POLLING_DELAY;
+
 enum avalon2_fan_fixed opt_avalon2_fan_fixed = FAN_AUTO;
 
-static struct pool pool_stratum;
+#define UNPACK32(x, str)			\
+{						\
+	*((str) + 3) = (uint8_t) ((x)      );	\
+	*((str) + 2) = (uint8_t) ((x) >>  8);	\
+	*((str) + 1) = (uint8_t) ((x) >> 16);	\
+	*((str) + 0) = (uint8_t) ((x) >> 24);	\
+}
+
+static void sha256_prehash(const unsigned char *message, unsigned int len, unsigned char *digest)
+{
+	sha256_ctx ctx;
+	int i;
+	sha256_init(&ctx);
+	sha256_update(&ctx, message, len);
+
+	for (i = 0; i < 8; i++) {
+		UNPACK32(ctx.h[i], &digest[i << 2]);
+	}
+}
 
 static inline uint8_t rev8(uint8_t d)
 {
-    int i;
-    uint8_t out = 0;
+	int i;
+	uint8_t out = 0;
 
-    /* (from left to right) */
-    for (i = 0; i < 8; i++)
-        if (d & (1 << i))
-            out |= (1 << (7 - i));
+	/* (from left to right) */
+	for (i = 0; i < 8; i++)
+		if (d & (1 << i))
+		out |= (1 << (7 - i));
 
-    return out;
+	return out;
 }
 
 char *set_avalon2_fan(char *arg)
@@ -84,8 +105,10 @@ char *set_avalon2_fan(char *arg)
 	if (val1 < 0 || val1 > 100 || val2 < 0 || val2 > 100 || val2 < val1)
 		return "Invalid value passed to avalon2-fan";
 
-	opt_avalon2_fan_min = get_fan_pwm(val1);
-	opt_avalon2_fan_max = get_fan_pwm(val2);
+	opt_avalon2_fan_min = val1;
+	opt_avalon2_fan_max = val2;
+	avalon2_fan_min = get_fan_pwm(val1);
+	avalon2_fan_max = get_fan_pwm(val2);
 
 	return NULL;
 }
@@ -165,7 +188,7 @@ static int job_idcmp(uint8_t *job_id, char *pool_job_id)
 		return 1;
 
 	job_id_len = strlen(pool_job_id);
-	crc_expect = crc16(pool_job_id, job_id_len);
+	crc_expect = crc16((unsigned char *)pool_job_id, job_id_len);
 
 	crc = job_id[0] << 8 | job_id[1];
 
@@ -215,39 +238,51 @@ static void adjust_fan(struct avalon2_info *info)
 {
 	int t;
 
-	if (unlikely(info->first < 2) || opt_avalon2_fan_fixed == FAN_FIXED) {
-		info->fan_pwm = opt_avalon2_fan_min;
+	if (opt_avalon2_fan_fixed == FAN_FIXED) {
+		info->fan_pct = opt_avalon2_fan_min;
+		info->fan_pwm = get_fan_pwm(info->fan_pct);
 		return;
 	}
 
 	t = get_current_temp_max(info);
 
 	/* TODO: Add options for temperature range and fan adjust function */
-	if (t < 50) {
-		info->fan_pwm = get_fan_pwm(40);
-		return;
-	}
+	if (t < 60)
+		info->fan_pct = opt_avalon2_fan_min;
+	else if (t > 80)
+		info->fan_pct = opt_avalon2_fan_max;
+	else
+		info->fan_pct = (t - 60) * (opt_avalon2_fan_max - opt_avalon2_fan_min) / 20 + opt_avalon2_fan_min;
 
-	if (t > 70) {
-		info->fan_pwm = get_fan_pwm(100);
-		return;
-	}
-
-	info->fan_pwm = get_fan_pwm(3 * t - 110);
+	info->fan_pwm = get_fan_pwm(info->fan_pct);
 }
 
 static inline int mm_cmp_1404(struct avalon2_info *info, int modular)
 {
+	/* <= 1404 return 1 */
 	char *mm_1404 = "1404";
 	return strncmp(info->mm_version[modular] + 2, mm_1404, 4) > 0 ? 0 : 1;
 }
 
-extern void submit_nonce2_nonce(struct thr_info *thr, struct pool *pool, struct pool *real_pool, uint32_t nonce2, uint32_t nonce);
+static inline int mm_cmp_1406(struct avalon2_info *info)
+{
+	/* <= 1406 return 1 */
+	char *mm_1406 = "1406";
+	int i;
+	for (i = 0; i < AVA2_DEFAULT_MODULARS; i++) {
+		if (info->enable[i] &&
+		    strncmp(info->mm_version[i] + 2, mm_1406, 4) <= 0)
+			return 1;
+	}
+
+	return 0;
+}
+
 static int decode_pkg(struct thr_info *thr, struct avalon2_ret *ar, uint8_t *pkg)
 {
-	struct cgpu_info *avalon2;
-	struct avalon2_info *info;
-	struct pool *pool, *real_pool;
+	struct cgpu_info *avalon2 = thr->cgpu;
+	struct avalon2_info *info = avalon2->device_data;
+	struct pool *pool, *real_pool, *pool_stratum = &info->pool;
 
 	unsigned int expected_crc;
 	unsigned int actual_crc;
@@ -258,11 +293,6 @@ static int decode_pkg(struct thr_info *thr, struct avalon2_ret *ar, uint8_t *pkg
 
 	int type = AVA2_GETS_ERROR;
 
-	if (thr) {
-		avalon2 = thr->cgpu;
-		info = avalon2->device_data;
-	}
-
 	memcpy((uint8_t *)ar, pkg, AVA2_READ_SIZE);
 
 	if (ar->head[0] == AVA2_H1 && ar->head[1] == AVA2_H2) {
@@ -271,7 +301,8 @@ static int decode_pkg(struct thr_info *thr, struct avalon2_ret *ar, uint8_t *pkg
 			((ar->crc[1] & 0xff) << 8);
 
 		type = ar->type;
-		applog(LOG_DEBUG, "Avalon2: %d: expected crc(%04x), actural_crc(%04x)", type, expected_crc, actual_crc);
+		applog(LOG_DEBUG, "Avalon2: %d: expected crc(%04x), actual_crc(%04x)",
+		       type, expected_crc, actual_crc);
 		if (expected_crc != actual_crc)
 			goto out;
 
@@ -282,6 +313,7 @@ static int decode_pkg(struct thr_info *thr, struct avalon2_ret *ar, uint8_t *pkg
 
 		switch(type) {
 		case AVA2_P_NONCE:
+			applog(LOG_DEBUG, "Avalon2: AVA2_P_NONCE");
 			memcpy(&miner, ar->data + 0, 4);
 			memcpy(&pool_no, ar->data + 4, 4);
 			memcpy(&nonce2, ar->data + 8, 4);
@@ -308,19 +340,19 @@ static int decode_pkg(struct thr_info *thr, struct avalon2_ret *ar, uint8_t *pkg
 
 			real_pool = pool = pools[pool_no];
 			if (job_idcmp(job_id, pool->swork.job_id)) {
-				if (!job_idcmp(job_id, pool_stratum.swork.job_id)) {
-					applog(LOG_ERR, "Avalon2: Match to previous stratum! (%s)", pool_stratum.swork.job_id);
-					pool = &pool_stratum;
+				if (!job_idcmp(job_id, pool_stratum->swork.job_id)) {
+					applog(LOG_DEBUG, "Avalon2: Match to previous stratum! (%s)", pool_stratum->swork.job_id);
+					pool = pool_stratum;
 				} else {
 					applog(LOG_ERR, "Avalon2: Cannot match to any stratum! (%s)", pool->swork.job_id);
 					break;
 				}
 			}
 
-			if (thr && info->first)
-				submit_nonce2_nonce(thr, pool, real_pool, nonce2, nonce);
+			submit_nonce2_nonce(thr, pool, real_pool, nonce2, nonce);
 			break;
 		case AVA2_P_STATUS:
+			applog(LOG_DEBUG, "Avalon2: AVA2_P_STATUS");
 			memcpy(&tmp, ar->data, 4);
 			tmp = be32toh(tmp);
 			info->temp[0 + modular_id * 2] = tmp >> 16;
@@ -353,12 +385,16 @@ static int decode_pkg(struct thr_info *thr, struct avalon2_ret *ar, uint8_t *pkg
 			avalon2->temp = get_temp_max(info);
 			break;
 		case AVA2_P_ACKDETECT:
+			applog(LOG_DEBUG, "Avalon2: AVA2_P_ACKDETECT");
 			break;
 		case AVA2_P_ACK:
+			applog(LOG_DEBUG, "Avalon2: AVA2_P_ACK");
 			break;
 		case AVA2_P_NAK:
+			applog(LOG_DEBUG, "Avalon2: AVA2_P_NAK");
 			break;
 		default:
+			applog(LOG_DEBUG, "Avalon2: Unknown response");
 			type = AVA2_GETS_ERROR;
 			break;
 		}
@@ -368,108 +404,73 @@ out:
 	return type;
 }
 
-static inline int avalon2_gets(int fd, uint8_t *buf)
+static inline int avalon2_gets(struct cgpu_info *avalon2, uint8_t *buf)
 {
-	int i;
-	int read_amount = AVA2_READ_SIZE;
-	uint8_t buf_tmp[AVA2_READ_SIZE];
-	uint8_t buf_copy[2 * AVA2_READ_SIZE];
+	int read_amount = AVA2_READ_SIZE, ret = 0;
 	uint8_t *buf_back = buf;
-	ssize_t ret = 0;
 
 	while (true) {
-		struct timeval timeout;
-		fd_set rd;
+		int err;
 
-		timeout.tv_sec = 0;
-		timeout.tv_usec = 100000;
-
-		FD_ZERO(&rd);
-		FD_SET(fd, &rd);
-		ret = select(fd + 1, &rd, NULL, NULL, &timeout);
-		if (unlikely(ret < 0)) {
-			applog(LOG_ERR, "Avalon2: Error %d on select in avalon_gets", errno);
-			return AVA2_GETS_ERROR;
-		}
-		if (ret) {
+		do {
 			memset(buf, 0, read_amount);
-			ret = read(fd, buf, read_amount);
-			if (unlikely(ret < 0)) {
-				applog(LOG_ERR, "Avalon2: Error %d on read in avalon_gets", errno);
+			err = usb_read(avalon2, (char *)buf, read_amount, &ret, C_AVA2_READ);
+			if (unlikely(err && err != LIBUSB_ERROR_TIMEOUT)) {
+				applog(LOG_ERR, "Avalon2: Error %d on read in avalon_gets got %d", err, ret);
 				return AVA2_GETS_ERROR;
 			}
 			if (likely(ret >= read_amount)) {
-				for (i = 1; i < read_amount; i++) {
-					if (buf_back[i - 1] == AVA2_H1 && buf_back[i] == AVA2_H2)
-						break;
-				}
-				i -= 1;
-				if (i) {
-					ret = read(fd, buf_tmp, i);
-					if (unlikely(ret != i)) {
-						applog(LOG_ERR, "Avalon2: Error %d on read in avalon_gets", errno);
-						return AVA2_GETS_ERROR;
-					}
-					memcpy(buf_copy, buf_back + i, AVA2_READ_SIZE - i);
-					memcpy(buf_copy + AVA2_READ_SIZE - i, buf_tmp, i);
-					memcpy(buf_back, buf_copy, AVA2_READ_SIZE);
-				}
+				if (unlikely(buf_back[0] != AVA2_H1 || buf_back[1] != AVA2_H2))
+					return AVA2_GETS_ERROR;
 				return AVA2_GETS_OK;
 			}
 			buf += ret;
 			read_amount -= ret;
-			continue;
-		}
+		} while (ret > 0);
 
 		return AVA2_GETS_TIMEOUT;
 	}
 }
 
-static int avalon2_send_pkg(int fd, const struct avalon2_pkg *pkg,
-			    struct thr_info __maybe_unused *thr)
+static int avalon2_send_pkg(struct cgpu_info *avalon2, const struct avalon2_pkg *pkg)
 {
-	int ret;
+	int err, amount;
 	uint8_t buf[AVA2_WRITE_SIZE];
 	int nr_len = AVA2_WRITE_SIZE;
 
+	if (unlikely(avalon2->usbinfo.nodev))
+		return AVA2_SEND_ERROR;
+
 	memcpy(buf, pkg, AVA2_WRITE_SIZE);
-	if (opt_debug) {
-		applog(LOG_DEBUG, "Avalon2: Sent(%d):", nr_len);
-		hexdump((uint8_t *)buf, nr_len);
-	}
-
-	ret = write(fd, buf, nr_len);
-	if (unlikely(ret != nr_len)) {
-		applog(LOG_DEBUG, "Avalon2: Send(%d)!", ret);
+	err = usb_write(avalon2, (char *)buf, nr_len, &amount, C_AVA2_WRITE);
+	if (err || amount != nr_len) {
+		applog(LOG_DEBUG, "Avalon2: Send(%d)!", amount);
 		return AVA2_SEND_ERROR;
 	}
 
-	cgsleep_ms(20);
-#if 0
-	ret = avalon2_gets(fd, result);
-	if (ret != AVA2_GETS_OK) {
-		applog(LOG_DEBUG, "Avalon2: Get(%d)!", ret);
-		return AVA2_SEND_ERROR;
-	}
-
-	ret = decode_pkg(thr, &ar, result);
-	if (ret != AVA2_P_ACK) {
-		applog(LOG_DEBUG, "Avalon2: PKG(%d)!", ret);
-		hexdump((uint8_t *)result, AVA2_READ_SIZE);
-		return AVA2_SEND_ERROR;
-	}
-#endif
-
 	return AVA2_SEND_OK;
 }
 
-static int avalon2_stratum_pkgs(int fd, struct pool *pool, struct thr_info *thr)
+static int avalon2_send_pkgs(struct cgpu_info *avalon2, const struct avalon2_pkg *pkg)
+{
+	int ret;
+
+	do {
+		if (unlikely(avalon2->usbinfo.nodev))
+			return -1;
+		ret = avalon2_send_pkg(avalon2, pkg);
+	} while (ret != AVA2_SEND_OK);
+
+	return 0;
+}
+
+static void avalon2_stratum_pkgs(struct cgpu_info *avalon2, struct pool *pool)
 {
 	const int merkle_offset = 36;
 	struct avalon2_pkg pkg;
 	int i, a, b, tmp;
 	unsigned char target[32];
-	int job_id_len;
+	int job_id_len, n2size;
 	unsigned short crc;
 
 	/* Send out the first stratum message STATIC */
@@ -486,7 +487,8 @@ static int avalon2_stratum_pkgs(int fd, struct pool *pool, struct thr_info *thr)
 	tmp = be32toh(pool->nonce2_offset);
 	memcpy(pkg.data + 4, &tmp, 4);
 
-	tmp = be32toh(pool->n2size);
+	n2size = pool->n2size >= 4 ? 4 : pool->n2size;
+	tmp = be32toh(n2size);
 	memcpy(pkg.data + 8, &tmp, 4);
 
 	tmp = be32toh(merkle_offset);
@@ -502,8 +504,8 @@ static int avalon2_stratum_pkgs(int fd, struct pool *pool, struct thr_info *thr)
 	memcpy(pkg.data + 24, &tmp, 4);
 
 	avalon2_init_pkg(&pkg, AVA2_P_STATIC, 1, 1);
-	while (avalon2_send_pkg(fd, &pkg, thr) != AVA2_SEND_OK)
-		;
+	if (avalon2_send_pkgs(avalon2, &pkg))
+		return;
 
 	set_target(target, pool->sdiff);
 	memcpy(pkg.data, target, 32);
@@ -514,47 +516,76 @@ static int avalon2_stratum_pkgs(int fd, struct pool *pool, struct thr_info *thr)
 		free(target_str);
 	}
 	avalon2_init_pkg(&pkg, AVA2_P_TARGET, 1, 1);
-	while (avalon2_send_pkg(fd, &pkg, thr) != AVA2_SEND_OK)
-		;
-
+	if (avalon2_send_pkgs(avalon2, &pkg))
+		return;
 
 	applog(LOG_DEBUG, "Avalon2: Pool stratum message JOBS_ID: %s",
 	       pool->swork.job_id);
 	memset(pkg.data, 0, AVA2_P_DATA_LEN);
 
 	job_id_len = strlen(pool->swork.job_id);
-	crc = crc16(pool->swork.job_id, job_id_len);
+	crc = crc16((unsigned char *)pool->swork.job_id, job_id_len);
 	pkg.data[0] = (crc & 0xff00) >> 8;
 	pkg.data[1] = crc & 0x00ff;
 	avalon2_init_pkg(&pkg, AVA2_P_JOB_ID, 1, 1);
-	while (avalon2_send_pkg(fd, &pkg, thr) != AVA2_SEND_OK)
-		;
+	if (avalon2_send_pkgs(avalon2, &pkg))
+		return;
 
-	a = pool->coinbase_len / AVA2_P_DATA_LEN;
-	b = pool->coinbase_len % AVA2_P_DATA_LEN;
-	applog(LOG_DEBUG, "Avalon2: Pool stratum message COINBASE: %d %d", a, b);
-	for (i = 0; i < a; i++) {
-		memcpy(pkg.data, pool->coinbase + i * 32, 32);
-		avalon2_init_pkg(&pkg, AVA2_P_COINBASE, i + 1, a + (b ? 1 : 0));
-		while (avalon2_send_pkg(fd, &pkg, thr) != AVA2_SEND_OK)
-			;
-	}
-	if (b) {
-		memset(pkg.data, 0, AVA2_P_DATA_LEN);
-		memcpy(pkg.data, pool->coinbase + i * 32, b);
-		avalon2_init_pkg(&pkg, AVA2_P_COINBASE, i + 1, i + 1);
-		while (avalon2_send_pkg(fd, &pkg, thr) != AVA2_SEND_OK)
-			;
+	if (pool->coinbase_len > AVA2_P_COINBASE_SIZE) {
+		int coinbase_len_posthash, coinbase_len_prehash;
+		uint8_t coinbase_prehash[32];
+		coinbase_len_prehash = pool->nonce2_offset - (pool->nonce2_offset % SHA256_BLOCK_SIZE);
+		coinbase_len_posthash = pool->coinbase_len - coinbase_len_prehash;
+		sha256_prehash(pool->coinbase, coinbase_len_prehash, coinbase_prehash);
+
+		a = (coinbase_len_posthash / AVA2_P_DATA_LEN) + 1;
+		b = coinbase_len_posthash % AVA2_P_DATA_LEN;
+	        memcpy(pkg.data, coinbase_prehash, 32);
+	        avalon2_init_pkg(&pkg, AVA2_P_COINBASE, 1, a + (b ? 1 : 0));
+	        if (avalon2_send_pkg(avalon2, &pkg))
+			return;
+	        applog(LOG_DEBUG, "Avalon2: Pool stratum message modified COINBASE: %d %d", a, b);
+	        for (i = 1; i < a; i++) {
+	                memcpy(pkg.data, pool->coinbase + coinbase_len_prehash + i * 32 - 32, 32);
+	                avalon2_init_pkg(&pkg, AVA2_P_COINBASE, i + 1, a + (b ? 1 : 0));
+			if (avalon2_send_pkg(avalon2, &pkg))
+	                        return;
+	        }
+	        if (b) {
+	                memset(pkg.data, 0, AVA2_P_DATA_LEN);
+			memcpy(pkg.data, pool->coinbase + coinbase_len_prehash + i * 32 - 32, b);
+	                avalon2_init_pkg(&pkg, AVA2_P_COINBASE, i + 1, i + 1);
+	                if (avalon2_send_pkg(avalon2, &pkg))
+	                        return;
+	        }
+	} else {
+		a = pool->coinbase_len / AVA2_P_DATA_LEN;
+		b = pool->coinbase_len % AVA2_P_DATA_LEN;
+		applog(LOG_DEBUG, "Avalon2: Pool stratum message COINBASE: %d %d", a, b);
+		for (i = 0; i < a; i++) {
+			memcpy(pkg.data, pool->coinbase + i * 32, 32);
+			avalon2_init_pkg(&pkg, AVA2_P_COINBASE, i + 1, a + (b ? 1 : 0));
+			if (avalon2_send_pkg(avalon2, &pkg))
+				return;
+		}
+		if (b) {
+			memset(pkg.data, 0, AVA2_P_DATA_LEN);
+			memcpy(pkg.data, pool->coinbase + i * 32, b);
+			avalon2_init_pkg(&pkg, AVA2_P_COINBASE, i + 1, i + 1);
+			if (avalon2_send_pkg(avalon2, &pkg))
+				return;
+		}
 	}
 
+
 	b = pool->merkles;
 	applog(LOG_DEBUG, "Avalon2: Pool stratum message MERKLES: %d", b);
 	for (i = 0; i < b; i++) {
 		memset(pkg.data, 0, AVA2_P_DATA_LEN);
 		memcpy(pkg.data, pool->swork.merkle_bin[i], 32);
 		avalon2_init_pkg(&pkg, AVA2_P_MERKLES, i + 1, b);
-		while (avalon2_send_pkg(fd, &pkg, thr) != AVA2_SEND_OK)
-			;
+		if (avalon2_send_pkgs(avalon2, &pkg))
+			return;
 	}
 
 	applog(LOG_DEBUG, "Avalon2: Pool stratum message HEADER: 4");
@@ -562,66 +593,67 @@ static int avalon2_stratum_pkgs(int fd, struct pool *pool, struct thr_info *thr)
 		memset(pkg.data, 0, AVA2_P_HEADER);
 		memcpy(pkg.data, pool->header_bin + i * 32, 32);
 		avalon2_init_pkg(&pkg, AVA2_P_HEADER, i + 1, 4);
-		while (avalon2_send_pkg(fd, &pkg, thr) != AVA2_SEND_OK)
-			;
-
+		if (avalon2_send_pkgs(avalon2, &pkg))
+			return;
 	}
-	return 0;
 }
 
-static int avalon2_get_result(struct thr_info *thr, int fd_detect, struct avalon2_ret *ar)
+static void avalon2_initialise(struct cgpu_info *avalon2)
 {
-	struct cgpu_info *avalon2;
-	struct avalon2_info *info;
-	int fd;
+	uint32_t ava2_data[2] = { PL2303_VALUE_LINE0, PL2303_VALUE_LINE1 };
+	int interface;
 
-	fd = fd_detect;
-	if (thr) {
-		avalon2 = thr->cgpu;
-		info = avalon2->device_data;
-		fd = info->fd;
-	}
+	if (avalon2->usbinfo.nodev)
+		return;
 
-	uint8_t result[AVA2_READ_SIZE];
-	int ret;
+	interface = usb_interface(avalon2);
+	// Set Data Control
+	usb_transfer(avalon2, PL2303_VENDOR_OUT, PL2303_REQUEST_VENDOR, 8,
+		     interface, C_VENDOR);
+	if (avalon2->usbinfo.nodev)
+		return;
 
-	memset(result, 0, AVA2_READ_SIZE);
+	usb_transfer(avalon2, PL2303_VENDOR_OUT, PL2303_REQUEST_VENDOR, 9,
+		     interface, C_VENDOR);
 
-	ret = avalon2_gets(fd, result);
-	if (ret != AVA2_GETS_OK)
-		return ret;
+	if (avalon2->usbinfo.nodev)
+		return;
 
-	if (opt_debug) {
-		applog(LOG_DEBUG, "Avalon2: Get(ret = %d):", ret);
-		hexdump((uint8_t *)result, AVA2_READ_SIZE);
-	}
+	// Set Line Control
+	usb_transfer_data(avalon2, PL2303_CTRL_OUT, PL2303_REQUEST_LINE, PL2303_VALUE_LINE,
+			  interface, ava2_data, PL2303_VALUE_LINE_SIZE, C_SETLINE);
+	if (avalon2->usbinfo.nodev)
+		return;
 
-	return decode_pkg(thr, ar, result);
+	// Vendor
+	usb_transfer(avalon2, PL2303_VENDOR_OUT, PL2303_REQUEST_VENDOR, PL2303_VALUE_VENDOR,
+		     interface, C_VENDOR);
+
+	if (avalon2->usbinfo.nodev)
+		return;
+
+	// Set More Line Control ?
+	usb_transfer(avalon2, PL2303_CTRL_OUT, PL2303_REQUEST_CTRL, 3, interface, C_SETLINE);
 }
 
-static bool avalon2_detect_one(const char *devpath)
+static struct cgpu_info *avalon2_detect_one(struct libusb_device *dev, struct usb_find_devices *found)
 {
 	struct avalon2_info *info;
 	int ackdetect;
-	int fd;
-	int tmp, i, j, modular[AVA2_DEFAULT_MODULARS];
+	int err, amount;
+	int tmp, i, j, modular[AVA2_DEFAULT_MODULARS] = {};
 	char mm_version[AVA2_DEFAULT_MODULARS][16];
 
-	struct cgpu_info *avalon2;
+	struct cgpu_info *avalon2 = usb_alloc_cgpu(&avalon2_drv, 1);
 	struct avalon2_pkg detect_pkg;
 	struct avalon2_ret ret_pkg;
 
-	applog(LOG_DEBUG, "Avalon2 Detect: Attempting to open %s", devpath);
-
-	fd = avalon2_open(devpath, AVA2_IO_SPEED, true);
-	if (unlikely(fd == -1)) {
-		applog(LOG_ERR, "Avalon2 Detect: Failed to open %s", devpath);
-		return false;
+	if (!usb_init(avalon2, dev, found)) {
+		applog(LOG_ERR, "Avalon2 failed usb_init");
+		avalon2 = usb_free_cgpu(avalon2);
+		return NULL;
 	}
-	tcflush(fd, TCIOFLUSH);
-
-	for (i = 0; i < AVA2_DEFAULT_MODULARS; i++)
-		modular[i] = 0;
+	avalon2_initialise(avalon2);
 
 	for (j = 0; j < 2; j++) {
 		for (i = 0; i < AVA2_DEFAULT_MODULARS; i++) {
@@ -632,8 +664,14 @@ static bool avalon2_detect_one(const char *devpath)
 			memcpy(detect_pkg.data + 28, &tmp, 4);
 
 			avalon2_init_pkg(&detect_pkg, AVA2_P_DETECT, 1, 1);
-			avalon2_send_pkg(fd, &detect_pkg, NULL);
-			ackdetect = avalon2_get_result(NULL, fd, &ret_pkg);
+			avalon2_send_pkg(avalon2, &detect_pkg);
+			err = usb_read(avalon2, (char *)&ret_pkg, AVA2_READ_SIZE, &amount, C_AVA2_READ);
+			if (err < 0 || amount != AVA2_READ_SIZE) {
+				applog(LOG_DEBUG, "%s %d: Avalon2 failed usb_read with err %d amount %d",
+				       avalon2->drv->name, avalon2->device_id, err, amount);
+				continue;
+			}
+			ackdetect = ret_pkg.type;
 			applog(LOG_DEBUG, "Avalon2 Detect ID[%d]: %d", i, ackdetect);
 			if (ackdetect != AVA2_P_ACKDETECT && modular[i] == 0)
 				continue;
@@ -642,31 +680,29 @@ static bool avalon2_detect_one(const char *devpath)
 			mm_version[i][15] = '\0';
 		}
 	}
-	avalon2_close(fd);
-	if (!modular[0] && !modular[1] && !modular[2] && !modular[3])
-		return false;
+	if (!modular[0] && !modular[1] && !modular[2] && !modular[3]) {
+		applog(LOG_DEBUG, "Not an Avalon2 device");
+		usb_uninit(avalon2);
+		usb_free_cgpu(avalon2);
+		return NULL;
+	}
 
 	/* We have a real Avalon! */
-	avalon2 = calloc(1, sizeof(struct cgpu_info));
-	avalon2->drv = &avalon2_drv;
-	avalon2->device_path = strdup(devpath);
 	avalon2->threads = AVA2_MINER_THREADS;
 	add_cgpu(avalon2);
 
-	applog(LOG_INFO, "Avalon2 Detect: Found at %s, mark as %d",
-	       devpath, avalon2->device_id);
+	update_usb_stats(avalon2);
+
+	applog(LOG_INFO, "%s%d: Found at %s", avalon2->drv->name, avalon2->device_id,
+	       avalon2->device_path);
 
 	avalon2->device_data = calloc(sizeof(struct avalon2_info), 1);
 	if (unlikely(!(avalon2->device_data)))
-		quit(1, "Failed to malloc avalon2_info");
+		quit(1, "Failed to calloc avalon2_info");
 
 	info = avalon2->device_data;
 
-	info->fd = -1;
-	info->baud = AVA2_IO_SPEED;
-	info->fan_pwm = AVA2_DEFAULT_FAN_PWM;
-	info->set_voltage = AVA2_DEFAULT_VOLTAGE_MIN;
-	info->set_frequency = AVA2_DEFAULT_FREQUENCY;
+	info->fan_pwm = get_fan_pwm(AVA2_DEFAULT_FAN_PWM);
 	info->temp_max = 0;
 
 	for (i = 0; i < AVA2_DEFAULT_MODULARS; i++) {
@@ -675,35 +711,29 @@ static bool avalon2_detect_one(const char *devpath)
 		info->enable[i] = modular[i];
 		info->dev_type[i] = AVA2_ID_AVAX;
 
-		if (!strncmp((char *)&(info->mm_version[i]), AVA2_FW2_PREFIXSTR, 2))
+		if (!strncmp((char *)&(info->mm_version[i]), AVA2_FW2_PREFIXSTR, 2)) {
 			info->dev_type[i] = AVA2_ID_AVA2;
-		if (!strncmp((char *)&(info->mm_version[i]), AVA2_FW3_PREFIXSTR, 2))
+			info->set_voltage = AVA2_DEFAULT_VOLTAGE_MIN;
+			info->set_frequency = AVA2_DEFAULT_FREQUENCY;
+		}
+		if (!strncmp((char *)&(info->mm_version[i]), AVA2_FW3_PREFIXSTR, 2)) {
 			info->dev_type[i] = AVA2_ID_AVA3;
+			info->set_voltage = AVA2_AVA3_VOLTAGE;
+			info->set_frequency = AVA2_AVA3_FREQUENCY;
+		}
 	}
 
-	pool_stratum.swork.job_id = NULL;
-	pool_stratum.merkles = 0;
-	return true;
+	if (!opt_avalon2_voltage_min)
+		opt_avalon2_voltage_min = opt_avalon2_voltage_max = info->set_voltage;
+	if (!opt_avalon2_freq_min)
+		opt_avalon2_freq_min = opt_avalon2_freq_max = info->set_frequency;
+
+	return avalon2;
 }
 
 static inline void avalon2_detect(bool __maybe_unused hotplug)
 {
-	serial_detect(&avalon2_drv, avalon2_detect_one);
-}
-
-static void avalon2_init(struct cgpu_info *avalon2)
-{
-	int fd;
-	struct avalon2_info *info = avalon2->device_data;
-
-	fd = avalon2_open(avalon2->device_path, info->baud, true);
-	if (unlikely(fd == -1)) {
-		applog(LOG_ERR, "Avalon2: Failed to open on %s", avalon2->device_path);
-		return;
-	}
-	applog(LOG_DEBUG, "Avalon2: Opened on %s", avalon2->device_path);
-
-	info->fd = fd;
+	usb_detect(&avalon2_drv, avalon2_detect_one);
 }
 
 static bool avalon2_prepare(struct thr_info *thr)
@@ -711,35 +741,23 @@ static bool avalon2_prepare(struct thr_info *thr)
 	struct cgpu_info *avalon2 = thr->cgpu;
 	struct avalon2_info *info = avalon2->device_data;
 
-	free(avalon2->works);
-	avalon2->works = calloc(sizeof(struct work *), 2);
-	if (!avalon2->works)
-		quit(1, "Failed to calloc avalon2 works in avalon2_prepare");
-
-	if (info->fd == -1)
-		avalon2_init(avalon2);
-
-	info->first = 0;
-
-	cglock_init(&(pool_stratum.data_lock));
+	cglock_init(&info->pool.data_lock);
 
 	return true;
 }
 
-static int polling(struct thr_info *thr)
+static int polling(struct thr_info *thr, struct cgpu_info *avalon2, struct avalon2_info *info)
 {
-	int i, tmp;
-
 	struct avalon2_pkg send_pkg;
 	struct avalon2_ret ar;
+	int i, tmp;
 
-	struct cgpu_info *avalon2 = thr->cgpu;
-	struct avalon2_info *info = avalon2->device_data;
-
-	static int pre_led_red[AVA2_DEFAULT_MODULARS];
 	for (i = 0; i < AVA2_DEFAULT_MODULARS; i++) {
 		if (info->modulars[i] && info->enable[i]) {
-			cgsleep_ms(20);
+			uint8_t result[AVA2_READ_SIZE];
+			int ret;
+
+			cgsleep_ms(opt_avalon2_polling_delay);
 			memset(send_pkg.data, 0, AVA2_P_DATA_LEN);
 
 			tmp = be32toh(info->led_red[i]); /* RED LED */
@@ -749,149 +767,174 @@ static int polling(struct thr_info *thr)
 			memcpy(send_pkg.data + 28, &tmp, 4);
 			if (info->led_red[i] && mm_cmp_1404(info, i)) {
 				avalon2_init_pkg(&send_pkg, AVA2_P_TEST, 1, 1);
-				while (avalon2_send_pkg(info->fd, &send_pkg, thr) != AVA2_SEND_OK)
-					;
+				avalon2_send_pkgs(avalon2, &send_pkg);
 				info->enable[i] = 0;
 				continue;
 			} else
 				avalon2_init_pkg(&send_pkg, AVA2_P_POLLING, 1, 1);
 
-			while (avalon2_send_pkg(info->fd, &send_pkg, thr) != AVA2_SEND_OK)
-				;
-			avalon2_get_result(thr, info->fd, &ar);
+			avalon2_send_pkgs(avalon2, &send_pkg);
+			ret = avalon2_gets(avalon2, result);
+			if (ret == AVA2_GETS_OK)
+				decode_pkg(thr, &ar, result);
 		}
 	}
 
 	return 0;
 }
 
-static void copy_pool_stratum(struct pool *pool)
+static void copy_pool_stratum(struct avalon2_info *info, struct pool *pool)
 {
 	int i;
 	int merkles = pool->merkles;
 	size_t coinbase_len = pool->coinbase_len;
+	struct pool *pool_stratum = &info->pool;
 
-	if (!job_idcmp(pool->swork.job_id, pool_stratum.swork.job_id))
+	if (!job_idcmp((unsigned char *)pool->swork.job_id, pool_stratum->swork.job_id))
 		return;
 
-	cg_wlock(&(pool_stratum.data_lock));
-	free(pool_stratum.swork.job_id);
-	free(pool_stratum.nonce1);
-	free(pool_stratum.coinbase);
+	cg_wlock(&pool_stratum->data_lock);
+	free(pool_stratum->swork.job_id);
+	free(pool_stratum->nonce1);
+	free(pool_stratum->coinbase);
 
 	align_len(&coinbase_len);
-	pool_stratum.coinbase = calloc(coinbase_len, 1);
-	if (unlikely(!pool_stratum.coinbase))
+	pool_stratum->coinbase = calloc(coinbase_len, 1);
+	if (unlikely(!pool_stratum->coinbase))
 		quit(1, "Failed to calloc pool_stratum coinbase in avalon2");
-	memcpy(pool_stratum.coinbase, pool->coinbase, coinbase_len);
+	memcpy(pool_stratum->coinbase, pool->coinbase, coinbase_len);
 
 
-	for (i = 0; i < pool_stratum.merkles; i++)
-		free(pool_stratum.swork.merkle_bin[i]);
+	for (i = 0; i < pool_stratum->merkles; i++)
+		free(pool_stratum->swork.merkle_bin[i]);
 	if (merkles) {
-		pool_stratum.swork.merkle_bin = realloc(pool_stratum.swork.merkle_bin,
+		pool_stratum->swork.merkle_bin = realloc(pool_stratum->swork.merkle_bin,
 						 sizeof(char *) * merkles + 1);
 		for (i = 0; i < merkles; i++) {
-			pool_stratum.swork.merkle_bin[i] = malloc(32);
-			if (unlikely(!pool_stratum.swork.merkle_bin[i]))
+			pool_stratum->swork.merkle_bin[i] = malloc(32);
+			if (unlikely(!pool_stratum->swork.merkle_bin[i]))
 				quit(1, "Failed to malloc pool_stratum swork merkle_bin");
-			memcpy(pool_stratum.swork.merkle_bin[i], pool->swork.merkle_bin[i], 32);
+			memcpy(pool_stratum->swork.merkle_bin[i], pool->swork.merkle_bin[i], 32);
 		}
 	}
 
-	pool_stratum.sdiff = pool->sdiff;
-	pool_stratum.coinbase_len = pool->coinbase_len;
-	pool_stratum.nonce2_offset = pool->nonce2_offset;
-	pool_stratum.n2size = pool->n2size;
-	pool_stratum.merkles = pool->merkles;
+	pool_stratum->sdiff = pool->sdiff;
+	pool_stratum->coinbase_len = pool->coinbase_len;
+	pool_stratum->nonce2_offset = pool->nonce2_offset;
+	pool_stratum->n2size = pool->n2size;
+	pool_stratum->merkles = pool->merkles;
 
-	pool_stratum.swork.job_id = strdup(pool->swork.job_id);
-	pool_stratum.nonce1 = strdup(pool->nonce1);
+	pool_stratum->swork.job_id = strdup(pool->swork.job_id);
+	pool_stratum->nonce1 = strdup(pool->nonce1);
 
-	memcpy(pool_stratum.ntime, pool->ntime, sizeof(pool_stratum.ntime));
-	memcpy(pool_stratum.header_bin, pool->header_bin, sizeof(pool_stratum.header_bin));
-	cg_wunlock(&(pool_stratum.data_lock));
+	memcpy(pool_stratum->ntime, pool->ntime, sizeof(pool_stratum->ntime));
+	memcpy(pool_stratum->header_bin, pool->header_bin, sizeof(pool_stratum->header_bin));
+	cg_wunlock(&pool_stratum->data_lock);
+}
+
+static void avalon2_update(struct cgpu_info *avalon2)
+{
+	struct avalon2_info *info = avalon2->device_data;
+	struct thr_info *thr = avalon2->thr[0];
+	struct avalon2_pkg send_pkg;
+	uint32_t tmp, range, start;
+	struct work *work;
+	struct pool *pool;
+
+	applog(LOG_DEBUG, "Avalon2: New stratum: restart: %d, update: %d",
+	       thr->work_restart, thr->work_update);
+	thr->work_update = false;
+	thr->work_restart = false;
+
+	work = get_work(thr, thr->id); /* Make sure pool is ready */
+	discard_work(work); /* Don't leak memory */
+
+	pool = current_pool();
+	if (!pool->has_stratum)
+		quit(1, "Avalon2: MM have to use stratum pool");
+
+	if (pool->coinbase_len > AVA2_P_COINBASE_SIZE) {
+		applog(LOG_INFO, "Avalon2: MM pool coinbase length(%d) is more than %d",
+		       pool->coinbase_len, AVA2_P_COINBASE_SIZE);
+		if (mm_cmp_1406(info)) {
+			applog(LOG_ERR, "Avalon2: MM version less then 1406");
+			return;
+		}
+		if ((pool->coinbase_len - pool->nonce2_offset + 64) > AVA2_P_COINBASE_SIZE) {
+			applog(LOG_ERR, "Avalon2: MM pool modified coinbase length(%d) is more than %d",
+			       pool->coinbase_len - pool->nonce2_offset + 64, AVA2_P_COINBASE_SIZE);
+			return;
+		}
+	}
+	if (pool->merkles > AVA2_P_MERKLES_COUNT) {
+		applog(LOG_ERR, "Avalon2: MM merkles have to less then %d", AVA2_P_MERKLES_COUNT);
+		return;
+	}
+	if (pool->n2size < 3) {
+		applog(LOG_ERR, "Avalon2: MM nonce2 size have to >= 3 (%d)", pool->n2size);
+		return;
+	}
+
+	cgtime(&info->last_stratum);
+	cg_rlock(&pool->data_lock);
+	info->pool_no = pool->pool_no;
+	copy_pool_stratum(info, pool);
+	avalon2_stratum_pkgs(avalon2, pool);
+	cg_runlock(&pool->data_lock);
+
+	/* Configuer the parameter from outside */
+	adjust_fan(info);
+	info->set_voltage = opt_avalon2_voltage_min;
+	info->set_frequency = opt_avalon2_freq_min;
+
+	/* Set the Fan, Voltage and Frequency */
+	memset(send_pkg.data, 0, AVA2_P_DATA_LEN);
+
+	tmp = be32toh(info->fan_pwm);
+	memcpy(send_pkg.data, &tmp, 4);
+
+	applog(LOG_INFO, "Avalon2: Temp max: %d, Cut off temp: %d",
+	       get_current_temp_max(info), opt_avalon2_overheat);
+	if (get_current_temp_max(info) >= opt_avalon2_overheat)
+		tmp = encode_voltage(0);
+	else
+		tmp = encode_voltage(info->set_voltage);
+	tmp = be32toh(tmp);
+	memcpy(send_pkg.data + 4, &tmp, 4);
+
+	tmp = be32toh(info->set_frequency);
+	memcpy(send_pkg.data + 8, &tmp, 4);
+
+	/* Configure the nonce2 offset and range */
+	if (pool->n2size == 3)
+		range = 0xffffff / (total_devices + 1);
+	else
+		range = 0xffffffff / (total_devices + 1);
+	start = range * (avalon2->device_id + 1);
+
+	tmp = be32toh(start);
+	memcpy(send_pkg.data + 12, &tmp, 4);
+
+	tmp = be32toh(range);
+	memcpy(send_pkg.data + 16, &tmp, 4);
+
+	/* Package the data */
+	avalon2_init_pkg(&send_pkg, AVA2_P_SET, 1, 1);
+	avalon2_send_pkgs(avalon2, &send_pkg);
 }
 
 static int64_t avalon2_scanhash(struct thr_info *thr)
 {
-	struct avalon2_pkg send_pkg;
 	struct timeval current_stratum;
-
-	struct pool *pool;
 	struct cgpu_info *avalon2 = thr->cgpu;
 	struct avalon2_info *info = avalon2->device_data;
-
 	int64_t h;
-	uint32_t tmp, range, start;
 	int i;
 
-	if (thr->work_restart || thr->work_update || !info->first) {
-		applog(LOG_DEBUG, "Avalon2: New stratum: restart: %d, update: %d, first: %d",
-		       thr->work_restart, thr->work_update, info->first);
-		thr->work_update = false;
-		thr->work_restart = false;
-		get_work(thr, thr->id); /* Make sure pool is ready */
-
-		pool = current_pool();
-		if (!pool->has_stratum)
-			quit(1, "Avalon2: Miner Manager have to use stratum pool");
-		if (pool->coinbase_len > AVA2_P_COINBASE_SIZE) {
-			applog(LOG_ERR, "Avalon2: Miner Manager pool coinbase length have to less then %d", AVA2_P_COINBASE_SIZE);
-			return 0;
-		}
-		if (pool->merkles > AVA2_P_MERKLES_COUNT) {
-			applog(LOG_ERR, "Avalon2: Miner Manager merkles have to less then %d", AVA2_P_MERKLES_COUNT);
-			return 0;
-		}
-
-		cgtime(&info->last_stratum);
-		cg_rlock(&pool->data_lock);
-		info->pool_no = pool->pool_no;
-		copy_pool_stratum(pool);
-		avalon2_stratum_pkgs(info->fd, pool, thr);
-		cg_runlock(&pool->data_lock);
-
-		/* Configuer the parameter from outside */
-		adjust_fan(info);
-		info->set_voltage = opt_avalon2_voltage_min;
-		info->set_frequency = opt_avalon2_freq_min;
-
-		/* Set the Fan, Voltage and Frequency */
-		memset(send_pkg.data, 0, AVA2_P_DATA_LEN);
-
-		tmp = be32toh(info->fan_pwm);
-		memcpy(send_pkg.data, &tmp, 4);
-
-		applog(LOG_ERR, "Avalon2: Temp max: %d, Cut off temp: %d",
-		       get_current_temp_max(info), opt_avalon2_overheat);
-		if (get_current_temp_max(info) >= opt_avalon2_overheat)
-			tmp = encode_voltage(0);
-		else
-			tmp = encode_voltage(info->set_voltage);
-		tmp = be32toh(tmp);
-		memcpy(send_pkg.data + 4, &tmp, 4);
-
-		tmp = be32toh(info->set_frequency);
-		memcpy(send_pkg.data + 8, &tmp, 4);
-
-		/* Configure the nonce2 offset and range */
-		range = 0xffffffff / total_devices;
-		start = range * avalon2->device_id;
-
-		tmp = be32toh(start);
-		memcpy(send_pkg.data + 12, &tmp, 4);
-
-		tmp = be32toh(range);
-		memcpy(send_pkg.data + 16, &tmp, 4);
-
-		/* Package the data */
-		avalon2_init_pkg(&send_pkg, AVA2_P_SET, 1, 1);
-		while (avalon2_send_pkg(info->fd, &send_pkg, thr) != AVA2_SEND_OK)
-			;
-
-		if (unlikely(info->first < 2))
-			info->first++;
+	if (unlikely(avalon2->usbinfo.nodev)) {
+		applog(LOG_ERR, "%s%d: Device disappeared, shutting down thread",
+		       avalon2->drv->name, avalon2->device_id);
+		return -1;
 	}
 
 	/* Stop polling the device if there is no stratum in 3 minutes, network is down */
@@ -899,7 +942,7 @@ static int64_t avalon2_scanhash(struct thr_info *thr)
 	if (tdiff(&current_stratum, &(info->last_stratum)) > (double)(3.0 * 60.0))
 		return 0;
 
-	polling(thr);
+	polling(thr, avalon2, info);
 
 	h = 0;
 	for (i = 0; i < AVA2_DEFAULT_MODULARS; i++) {
@@ -921,7 +964,7 @@ static struct api_data *avalon2_api_stats(struct cgpu_info *cgpu)
 		if(info->dev_type[i] == AVA2_ID_AVAX)
 			continue;
 		sprintf(buf, "ID%d MM Version", i + 1);
-		root = api_add_string(root, buf, &(info->mm_version[i]), false);
+		root = api_add_string(root, buf, (char *)&(info->mm_version[i]), false);
 	}
 
 	minerindex = 0;
@@ -997,56 +1040,32 @@ static struct api_data *avalon2_api_stats(struct cgpu_info *cgpu)
 		sprintf(buf, "Power good %02x", i + 1);
 		root = api_add_int(root, buf, &(info->power_good[i]), false);
 	}
+	for (i = 0; i < AVA2_DEFAULT_MODULARS; i++) {
+		if(info->dev_type[i] == AVA2_ID_AVAX)
+			continue;
+		sprintf(buf, "Led %02x", i + 1);
+		root = api_add_int(root, buf, &(info->led_red[i]), false);
+	}
 
 	return root;
 }
 
-static char *avalon2_set_device(struct cgpu_info *avalon2, char *option, char *setting, char *replybuf)
+static void avalon2_statline_before(char *buf, size_t bufsiz, struct cgpu_info *avalon2)
 {
-	int val;
-	struct avalon2_info *info;
+	struct avalon2_info *info = avalon2->device_data;
+	int temp = get_current_temp_max(info);
+	float volts = (float)info->set_voltage / 10000;
 
-	if (strcasecmp(option, "help") == 0) {
-		sprintf(replybuf, "led: module_id");
-		return replybuf;
-	}
-
-	if (strcasecmp(option, "led") == 0) {
-		if (!setting || !*setting) {
-			sprintf(replybuf, "missing module_id setting");
-			return replybuf;
-		}
-
-		val = atoi(setting);
-		if (val < 1 || val > AVA2_DEFAULT_MODULARS) {
-			sprintf(replybuf, "invalid module_id: %d, valid range 1-%d", val, AVA2_DEFAULT_MODULARS);
-			return replybuf;
-		}
-
-		val -= 1;
-
-		info = avalon2->device_data;
-		info->led_red[val] = !info->led_red[val];
-		if (!info->led_red[val] && mm_cmp_1404(info, val)) {
-			applog(LOG_ERR, "Avalon2: MM early then MM_XX1404, enable module:%d", val + 1);
-			info->enable[val] = 1;
-		}
-
-		applog(LOG_ERR, "Avalon2: Module:%d, LED: %s", val + 1, info->led_red[val] ? "on" : "off");
-		return NULL;
-	}
-	/* TODO: Add other commands */
-
-	sprintf(replybuf, "Unknown option: %s", option);
-	return replybuf;
+	tailsprintf(buf, bufsiz, "%4dMhz %2dC %3d%% %.3fV", info->set_frequency,
+		    temp, info->fan_pct, volts);
 }
 
 static void avalon2_shutdown(struct thr_info *thr)
 {
-	struct cgpu_info *avalon = thr->cgpu;
+	struct cgpu_info *avalon2 = thr->cgpu;
+	int interface = usb_interface(avalon2);
 
-	free(avalon->works);
-	avalon->works = NULL;
+	usb_transfer(avalon2, PL2303_CTRL_OUT, PL2303_REQUEST_CTRL, 0, interface, C_SETLINE);
 }
 
 struct device_drv avalon2_drv = {
@@ -1054,11 +1073,12 @@ struct device_drv avalon2_drv = {
 	.dname = "avalon2",
 	.name = "AV2",
 	.get_api_stats = avalon2_api_stats,
-	.set_device = avalon2_set_device,
+	.get_statline_before = avalon2_statline_before,
 	.drv_detect = avalon2_detect,
-	.reinit_device = avalon2_init,
 	.thread_prepare = avalon2_prepare,
 	.hash_work = hash_driver_work,
+	.flush_work = avalon2_update,
+	.update_work = avalon2_update,
 	.scanwork = avalon2_scanhash,
 	.thread_shutdown = avalon2_shutdown,
 };
