@@ -1,5 +1,5 @@
 /*
- * Copyright 2013 Con Kolivas
+ * Copyright 2013-2014 Con Kolivas
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -16,13 +16,68 @@
 #include "libbitfury.h"
 
 int opt_bxf_temp_target = BXF_TEMP_TARGET / 10;
-int opt_nf1_bits = 50;
+int opt_nfu_bits = 50;
+int opt_bxm_bits = 54;
+int opt_bxf_bits = 54;
+int opt_bxf_debug;
+int opt_osm_led_mode = 4;
 
 /* Wait longer 1/3 longer than it would take for a full nonce range */
 #define BF1WAIT 1600
 #define BF1MSGSIZE 7
 #define BF1INFOSIZE 14
 
+#define TWELVE_MHZ 12000000
+
+//Low port pins
+#define SK      1
+#define DO      2
+#define DI      4
+#define CS      8
+#define GPIO0   16
+#define GPIO1   32
+#define GPIO2   64
+#define GPIO3   128
+
+//GPIO pins
+#define GPIOL0  0
+#define GPIOL1  1
+#define GPIOL2  2
+#define GPIOL3  3
+#define GPIOH   4
+#define GPIOH1  5
+#define GPIOH2  6
+#define GPIOH3  7
+#define GPIOH4  8
+#define GPIOH5  9
+#define GPIOH6  10
+#define GPIOH7  11
+
+#define DEFAULT_DIR            (SK | DO | CS | GPIO0 | GPIO1 | GPIO2 | GPIO3)  /* Setup default input or output state per FTDI for SPI */
+#define DEFAULT_STATE          (CS)                                       /* CS idles high, CLK idles LOW for SPI0 */
+
+//MPSSE commands from FTDI AN_108
+#define INVALID_COMMAND           0xAB
+#define ENABLE_ADAPTIVE_CLOCK     0x96
+#define DISABLE_ADAPTIVE_CLOCK    0x97
+#define ENABLE_3_PHASE_CLOCK      0x8C
+#define DISABLE_3_PHASE_CLOCK     0x8D
+#define TCK_X5                    0x8A
+#define TCK_D5                    0x8B
+#define CLOCK_N_CYCLES            0x8E
+#define CLOCK_N8_CYCLES           0x8F
+#define PULSE_CLOCK_IO_HIGH       0x94
+#define PULSE_CLOCK_IO_LOW        0x95
+#define CLOCK_N8_CYCLES_IO_HIGH   0x9C
+#define CLOCK_N8_CYCLES_IO_LOW    0x9D
+#define TRISTATE_IO               0x9E
+#define TCK_DIVISOR               0x86
+#define LOOPBACK_END              0x85
+#define SET_OUT_ADBUS             0x80
+#define SET_OUT_ACBUS             0x82
+#define WRITE_BYTES_SPI0          0x11
+#define READ_WRITE_BYTES_SPI0     0x31
+
 static void bf1_empty_buffer(struct cgpu_info *bitfury)
 {
 	char buf[512];
@@ -79,6 +134,7 @@ static void bitfury_identify(struct cgpu_info *bitfury)
 			bf1_identify(bitfury);
 			break;
 		case IDENT_BXF:
+		case IDENT_OSM:
 		default:
 			break;
 	}
@@ -109,9 +165,10 @@ static bool bf1_getinfo(struct cgpu_info *bitfury, struct bitfury_info *info)
 	info->version = buf[1];
 	memcpy(&info->product, buf + 2, 8);
 	memcpy(&info->serial, buf + 10, 4);
+	bitfury->unique_id = bin2hex((unsigned char *)buf + 10, 4);
 
-	applog(LOG_INFO, "%s %d: Getinfo returned version %d, product %s serial %08x", bitfury->drv->name,
-	       bitfury->device_id, info->version, info->product, info->serial);
+	applog(LOG_INFO, "%s %d: Getinfo returned version %d, product %s serial %s", bitfury->drv->name,
+	       bitfury->device_id, info->version, info->product, bitfury->unique_id);
 	bf1_empty_buffer(bitfury);
 	return true;
 }
@@ -152,6 +209,13 @@ static bool bxf_send_msg(struct cgpu_info *bitfury, char *buf, enum usb_cmds cmd
 	if (unlikely(bitfury->usbinfo.nodev))
 		return false;
 
+	if (opt_bxf_debug) {
+		char *strbuf = str_text(buf);
+
+		applog(LOG_ERR, "%s %d: >BXF [%s]", bitfury->drv->name, bitfury->device_id, strbuf);
+		free(strbuf);
+	}
+
 	len = strlen(buf);
 	applog(LOG_DEBUG, "%s %d: Sending %s", bitfury->drv->name, bitfury->device_id, buf);
 	err = usb_write(bitfury, buf, len, &amount, cmd);
@@ -163,6 +227,22 @@ static bool bxf_send_msg(struct cgpu_info *bitfury, char *buf, enum usb_cmds cmd
 	return true;
 }
 
+static bool bxf_send_debugmode(struct cgpu_info *bitfury)
+{
+	char buf[16];
+
+	sprintf(buf, "debug-mode %d\n", opt_bxf_debug);
+	return bxf_send_msg(bitfury, buf, C_BXF_DEBUGMODE);
+}
+
+static bool bxf_send_ledmode(struct cgpu_info *bitfury)
+{
+	char buf[16];
+
+	sprintf(buf, "led-mode %d\n", opt_osm_led_mode);
+	return bxf_send_msg(bitfury, buf, C_BXF_LEDMODE);
+}
+
 /* Returns the amount received only if we receive a full message, otherwise
  * it returns the err value. */
 static int bxf_recv_msg(struct cgpu_info *bitfury, char *buf)
@@ -238,6 +318,18 @@ static bool bxf_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
 	applog(LOG_INFO, "%s %d: Successfully initialised %s",
 	       bitfury->drv->name, bitfury->device_id, bitfury->device_path);
 
+	/* Sanity check and recognise variations */
+	if (info->chips <= 2 || info->chips > 999)
+		info->chips = 2;
+	else if (info->chips <= 6 && info->ident == IDENT_BXF)
+		bitfury->drv->name = "HXF";
+	else if (info->chips > 6 && info->ident == IDENT_BXF)
+		bitfury->drv->name = "MXF";
+	info->filtered_hw = calloc(sizeof(int), info->chips);
+	info->job = calloc(sizeof(int), info->chips);
+	info->submits = calloc(sizeof(int), info->chips);
+	if (!info->filtered_hw || !info->job || !info->submits)
+		quit(1, "Failed to calloc bxf chip arrays");
 	info->total_nonces = 1;
 	info->temp_target = opt_bxf_temp_target * 10;
 	/* This unsets it to make sure it gets set on the first pass */
@@ -279,7 +371,7 @@ out_close:
 	return false;
 }
 
-static void nf1_close(struct cgpu_info *bitfury)
+static void nfu_close(struct cgpu_info *bitfury)
 {
 	struct bitfury_info *info = bitfury->device_data;
 	struct mcp_settings *mcp = &info->mcp;
@@ -295,18 +387,27 @@ static void nf1_close(struct cgpu_info *bitfury)
 	mcp2210_set_gpio_settings(bitfury, mcp);
 }
 
-static bool nf1_reinit(struct cgpu_info *bitfury, struct bitfury_info *info)
+static bool nfu_reinit(struct cgpu_info *bitfury, struct bitfury_info *info)
 {
-	spi_clear_buf(info);
-	spi_add_break(info);
-	spi_set_freq(info);
-	spi_send_conf(info);
-	spi_send_init(info);
-	spi_reset(bitfury, info);
-	return spi_txrx(bitfury, info);
+	bool ret = true;
+	int i;
+
+	for (i = 0; i < info->chips; i++) {
+		spi_clear_buf(info);
+		spi_add_break(info);
+		spi_add_fasync(info, i);
+		spi_set_freq(info);
+		spi_send_conf(info);
+		spi_send_init(info);
+		spi_reset(bitfury, info);
+		ret = info->spi_txrx(bitfury, info);
+		if (!ret)
+			break;
+	}
+	return ret;
 }
 
-static bool nf1_set_spi_settings(struct cgpu_info *bitfury, struct bitfury_info *info)
+static bool nfu_set_spi_settings(struct cgpu_info *bitfury, struct bitfury_info *info)
 {
 	struct mcp_settings *mcp = &info->mcp;
 
@@ -314,7 +415,18 @@ static bool nf1_set_spi_settings(struct cgpu_info *bitfury, struct bitfury_info
 		mcp->acsv, mcp->cstdd, mcp->ldbtcsd, mcp->sdbd, mcp->bpst, mcp->spimode);
 }
 
-static bool nf1_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
+static void nfu_alloc_arrays(struct bitfury_info *info)
+{
+	info->payload = calloc(sizeof(struct bitfury_payload), info->chips);
+	info->oldbuf = calloc(sizeof(unsigned int) * 17, info->chips);
+	info->job_switched = calloc(sizeof(bool), info->chips);
+	info->second_run = calloc(sizeof(bool), info->chips);
+	info->work = calloc(sizeof(struct work *), info->chips);
+	info->owork = calloc(sizeof(struct work *), info->chips);
+	info->submits = calloc(sizeof(int *), info->chips);
+}
+
+static bool nfu_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
 {
 	struct mcp_settings *mcp = &info->mcp;
 	char buf[MCP2210_BUFFER_LENGTH];
@@ -322,6 +434,18 @@ static bool nf1_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
 	bool ret = false;
 	int i, val;
 
+	/* Identify number of chips, and use it in device name if it can fit
+	 * into 3 chars, otherwise use generic NFU name. */
+	val = sscanf(bitfury->usbdev->prod_string, "NanoFury NF%u ", &info->chips);
+	if (val < 1)
+		info->chips = 1;
+	else if (info->chips < 10) {
+		sprintf(info->product, "NF%u", info->chips);
+		bitfury->drv->name = info->product;
+	}
+	nfu_alloc_arrays(info);
+
+	info->spi_txrx = &mcp_spi_txrx;
 	mcp2210_get_gpio_settings(bitfury, mcp);
 
 	for (i = 0; i < 9; i++) {
@@ -333,8 +457,10 @@ static bool nf1_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
 	}
 
 	/* Set LED and PWR pins to output and high */
-	mcp->direction.pin[NF1_PIN_LED] = mcp->direction.pin[NF1_PIN_PWR_EN] = MCP2210_GPIO_OUTPUT;
-	mcp->value.pin[NF1_PIN_LED] = mcp->value.pin[NF1_PIN_PWR_EN] = MCP2210_GPIO_PIN_HIGH;
+	mcp->direction.pin[NFU_PIN_LED] = mcp->direction.pin[NFU_PIN_PWR_EN] = MCP2210_GPIO_OUTPUT;
+	mcp->value.pin[NFU_PIN_LED] = mcp->value.pin[NFU_PIN_PWR_EN] = MCP2210_GPIO_PIN_HIGH;
+	mcp->direction.pin[NFU_PIN_PWR_EN0] = MCP2210_GPIO_OUTPUT;
+	mcp->value.pin[NFU_PIN_PWR_EN0] = MCP2210_GPIO_PIN_LOW;
 
 	mcp->direction.pin[4] = MCP2210_GPIO_OUTPUT;
 	mcp->designation.pin[4] = MCP2210_PIN_CS;
@@ -373,7 +499,7 @@ static bool nf1_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
 	mcp->acsv = 0xffef;
 	mcp->cstdd = mcp->ldbtcsd = mcp->sdbd = mcp->spimode = 0;
 	mcp->bpst = 1;
-	if (!nf1_set_spi_settings(bitfury, info))
+	if (!nfu_set_spi_settings(bitfury, info))
 		goto out;
 
 	buf[0] = 0;
@@ -382,14 +508,14 @@ static bool nf1_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
 		goto out;
 	/* after this command SCK_OVRRIDE should read the same as current SCK
 	 * value (which for mode 0 should be 0) */
-	if (!mcp2210_get_gpio_pinval(bitfury, NF1_PIN_SCK_OVR, &val))
+	if (!mcp2210_get_gpio_pinval(bitfury, NFU_PIN_SCK_OVR, &val))
 		goto out;
 	if (val != MCP2210_GPIO_PIN_LOW)
 		goto out;
 
 	/* switch SCK to polarity (default SCK=1 in mode 2) */
 	mcp->spimode = 2;
-	if (!nf1_set_spi_settings(bitfury, info))
+	if (!nfu_set_spi_settings(bitfury, info))
 		goto out;
 	buf[0] = 0;
 	length = 1;
@@ -397,42 +523,292 @@ static bool nf1_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
 		goto out;
 	/* after this command SCK_OVRRIDE should read the same as current SCK
 	 * value (which for mode 2 should be 1) */
-	if (!mcp2210_get_gpio_pinval(bitfury, NF1_PIN_SCK_OVR, &val))
+	if (!mcp2210_get_gpio_pinval(bitfury, NFU_PIN_SCK_OVR, &val))
 		goto out;
 	if (val != MCP2210_GPIO_PIN_HIGH)
 		goto out;
 
 	/* switch SCK to polarity (default SCK=0 in mode 0) */
 	mcp->spimode = 0;
-	if (!nf1_set_spi_settings(bitfury, info))
+	if (!nfu_set_spi_settings(bitfury, info))
 		goto out;
 	buf[0] = 0;
 	length = 1;
 	if (!mcp2210_spi_transfer(bitfury, mcp, buf, &length))
 		goto out;
-	if (!mcp2210_get_gpio_pinval(bitfury, NF1_PIN_SCK_OVR, &val))
+	if (!mcp2210_get_gpio_pinval(bitfury, NFU_PIN_SCK_OVR, &val))
 		goto out;
 	if (val != MCP2210_GPIO_PIN_LOW)
 		goto out;
 
-	info->osc6_bits = opt_nf1_bits;
-	if (!nf1_reinit(bitfury, info))
+	info->osc6_bits = opt_nfu_bits;
+	if (!nfu_reinit(bitfury, info))
 		goto out;
 
 	ret = true;
 	if (!add_cgpu(bitfury))
-		quit(1, "Failed to add_cgpu in nf1_detect_one");
+		quit(1, "Failed to add_cgpu in nfu_detect_one");
 
 	update_usb_stats(bitfury);
 	applog(LOG_INFO, "%s %d: Successfully initialised %s",
-	bitfury->drv->name, bitfury->device_id, bitfury->device_path);
+	       bitfury->drv->name, bitfury->device_id, bitfury->device_path);
+	spi_clear_buf(info);
+
+	info->total_nonces = info->chips;
+out:
+	if (!ret)
+		nfu_close(bitfury);
+
+	return ret;
+}
+
+static bool bxm_purge_buffers(struct cgpu_info *bitfury)
+{
+	int err;
+
+	err = usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_RESET_REQUEST, SIO_RESET_PURGE_RX, 1, C_BXM_PURGERX);
+	if (err)
+		return false;
+	err = usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_RESET_REQUEST, SIO_RESET_PURGE_TX, 1, C_BXM_PURGETX);
+	if (err)
+		return false;
+	return true;
+}
+
+/* Calculate required divisor for desired frequency see FTDI AN_108 page 19*/
+static uint16_t calc_divisor(uint32_t system_clock, uint32_t freq)
+{
+	uint16_t divisor = system_clock / freq;
+
+	divisor /= 2;
+	divisor -= 1;
+	return divisor;
+}
+
+static void bxm_shutdown(struct cgpu_info *bitfury, struct bitfury_info *info)
+{
+	int chip_n;
+
+	for (chip_n = 0; chip_n < 2; chip_n++) {
+		spi_clear_buf(info);
+		spi_add_break(info);
+		spi_add_fasync(info, chip_n);
+		spi_config_reg(info, 4, 0);
+		info->spi_txrx(bitfury, info);
+	}
+}
+
+static void bxm_close(struct cgpu_info *bitfury, struct bitfury_info *info)
+{
+	unsigned char bitmask = 0;
+	unsigned char mode = BITMODE_RESET;
+	unsigned short usb_val = bitmask;
+
+	bxm_shutdown(bitfury, info);
+
+	//Need to do BITMODE_RESET before usb close per FTDI
+	usb_val |= (mode << 8);
+	usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_SET_BITMODE_REQUEST, usb_val, 1, C_BXM_SETBITMODE);
+}
+
+static bool bxm_open(struct cgpu_info *bitfury)
+{
+	unsigned char mode = BITMODE_RESET;
+	unsigned char bitmask = 0;
+	unsigned short usb_val = bitmask;
+	uint32_t system_clock = TWELVE_MHZ;
+	uint32_t freq = 200000;
+	uint16_t divisor = calc_divisor(system_clock,freq);
+	int amount, err;
+	char buf[4];
+
+	/* Enable the transaction translator emulator for these devices
+	 * otherwise we may write to them too quickly. */
+	bitfury->usbdev->tt = true;
+
+	err = usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_RESET_REQUEST, SIO_RESET_SIO, 1, C_BXM_SRESET);
+	if (err)
+		return false;
+	err = usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_SET_LATENCY_TIMER_REQUEST, BXM_LATENCY_MS, 1, C_BXM_SETLATENCY);
+	if (err)
+		return false;
+	err = usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_SET_EVENT_CHAR_REQUEST, 0x00, 1, C_BXM_SECR);
+	if (err)
+		return false;
+
+	//Do a BITMODE_RESET
+	usb_val |= (mode << 8);
+	err = usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_SET_BITMODE_REQUEST, usb_val, 1, C_BXM_SETBITMODE);
+	if (err)
+		return false;
+	//Now set to MPSSE mode
+	bitmask = 0;
+	mode = BITMODE_MPSSE;
+	usb_val = bitmask;
+	usb_val |= (mode << 8);
+	err = usb_transfer(bitfury, FTDI_TYPE_OUT, SIO_SET_BITMODE_REQUEST, usb_val, 1, C_BXM_SETBITMODE);
+	if (err)
+		return false;
+
+	//Now set the clock divisor
+	//First send just the 0x8B command to set the system clock to 12MHz
+	memset(buf, 0, 4);
+	buf[0] = TCK_D5;
+	err = usb_write(bitfury, buf, 1, &amount, C_BXM_CLOCK);
+	if (err || amount != 1)
+		return false;
+
+	buf[0] = TCK_DIVISOR;
+	buf[1] = (divisor & 0xFF);
+	buf[2] = ((divisor >> 8) & 0xFF);
+	err = usb_write(bitfury, buf, 3, &amount, C_BXM_CLOCKDIV);
+	if (err || amount != 3)
+		return false;
+
+	//Disable internal loopback
+	buf[0] = LOOPBACK_END;
+	err = usb_write(bitfury, buf, 1, &amount, C_BXM_LOOP);
+	if (err || amount != 1)
+		return false;
+
+	//Now set direction and idle (initial) states for the pins
+	buf[0] = SET_OUT_ADBUS;
+	buf[1] = DEFAULT_STATE; //Bitmask for LOW_PORT
+	buf[2] = DEFAULT_DIR;
+	err = usb_write(bitfury, buf, 3, &amount, C_BXM_ADBUS);
+	if (err || amount != 3)
+		return false;
+
+	//Set the pin states for the HIGH_BITS port as all outputs, all low
+	buf[0] = SET_OUT_ACBUS;
+	buf[1] = 0x00; //Bitmask for HIGH_PORT
+	buf[2] = 0xFF;
+	err = usb_write(bitfury, buf, 3, &amount, C_BXM_ACBUS);
+	if (err || amount != 3)
+		return false;
+
+	return true;
+}
+
+static bool bxm_set_CS_low(struct cgpu_info *bitfury)
+{
+	char buf[4] = { 0 };
+	int err, amount;
+
+	buf[0] = SET_OUT_ADBUS;
+	buf[1] &= ~DEFAULT_STATE; //Bitmask for LOW_PORT
+	buf[2] = DEFAULT_DIR;
+	err = usb_write(bitfury, buf, 3, &amount, C_BXM_CSLOW);
+	if (err || amount != 3)
+		return false;
+
+	return true;
+}
+
+static bool bxm_set_CS_high(struct cgpu_info *bitfury)
+{
+	char buf[4] = { 0 };
+	int err, amount;
+
+	buf[0] = SET_OUT_ADBUS;
+	buf[1] = DEFAULT_STATE; //Bitmask for LOW_PORT
+	buf[2] = DEFAULT_DIR;
+	err = usb_write(bitfury, buf, 3, &amount, C_BXM_CSHIGH);
+	if (err || amount != 3)
+		return false;
+
+	return true;
+}
+
+static bool bxm_reset_bitfury(struct cgpu_info *bitfury)
+{
+	char buf[20] = { 0 };
+	char rst_buf[8] = {0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0x00};
+	int err, amount;
+
+	//Set the FTDI CS pin HIGH. This will gate the clock to the Bitfury chips so we can send the reset sequence.
+	if (!bxm_set_CS_high(bitfury))
+		return false;
+
+	buf[0] = WRITE_BYTES_SPI0;
+	buf[1] = (uint8_t)16 - (uint8_t)1;
+	buf[2] = 0;
+	memcpy(&buf[3], rst_buf, 8);
+	memcpy(&buf[11], rst_buf, 8);
+	err = usb_write(bitfury, buf, 19, &amount, C_BXM_RESET);
+	if (err || amount != 19)
+		return false;
+
+	if (!bxm_set_CS_low(bitfury))
+		return false;
+
+	return true;
+}
+
+static bool bxm_reinit(struct cgpu_info *bitfury, struct bitfury_info *info)
+{
+	bool ret;
+	int i;
+
+	for (i = 0; i < 2; i++) {
+		spi_clear_buf(info);
+		spi_add_break(info);
+		spi_add_fasync(info, i);
+		spi_set_freq(info);
+		spi_send_conf(info);
+		spi_send_init(info);
+		ret = info->spi_txrx(bitfury, info);
+		if (!ret)
+			break;
+	}
+	return ret;
+}
+
+static bool bxm_detect_one(struct cgpu_info *bitfury, struct bitfury_info *info)
+{
+	bool ret;
+
+	info->spi_txrx = &ftdi_spi_txrx;
+	ret = bxm_open(bitfury);
+	if (!ret)
+		goto out;
+	ret = bxm_purge_buffers(bitfury);
+	if (!ret)
+		goto out;
+	ret = bxm_reset_bitfury(bitfury);
+	if (!ret)
+		goto out;
+	ret = bxm_purge_buffers(bitfury);
+	if (!ret)
+		goto out;
+
+	/* Do a dummy read */
+	memset(info->spibuf, 0, 80);
+	info->spibufsz = 80;
+	ret = info->spi_txrx(bitfury, info);
+	if (!ret)
+		goto out;
+	info->osc6_bits = opt_bxm_bits;
+	/* Only have 2 chip devices for now */
+	info->chips = 2;
+	nfu_alloc_arrays(info);
+
+	ret = bxm_reinit(bitfury, info);
+	if (!ret)
+		goto out;
+
+	if (!add_cgpu(bitfury))
+		quit(1, "Failed to add_cgpu in bxm_detect_one");
+
+	update_usb_stats(bitfury);
+	applog(LOG_INFO, "%s %d: Successfully initialised %s",
+	       bitfury->drv->name, bitfury->device_id, bitfury->device_path);
 	spi_clear_buf(info);
 
 	info->total_nonces = 1;
 out:
 	if (!ret)
-		nf1_close(bitfury);
-
+		bxm_close(bitfury, info);
 	return ret;
 }
 
@@ -460,10 +836,14 @@ static struct cgpu_info *bitfury_detect_one(struct libusb_device *dev, struct us
 			ret = bf1_detect_one(bitfury, info);
 			break;
 		case IDENT_BXF:
+		case IDENT_OSM:
 			ret = bxf_detect_one(bitfury, info);
 			break;
-		case IDENT_NF1:
-			ret = nf1_detect_one(bitfury, info);
+		case IDENT_NFU:
+			ret = nfu_detect_one(bitfury, info);
+			break;
+		case IDENT_BXM:
+			ret = bxm_detect_one(bitfury, info);
 			break;
 		default:
 			applog(LOG_INFO, "%s %d: Unrecognised bitfury device",
@@ -485,6 +865,30 @@ static void bitfury_detect(bool __maybe_unused hotplug)
 	usb_detect(&bitfury_drv, bitfury_detect_one);
 }
 
+static void adjust_bxf_chips(struct cgpu_info *bitfury, struct bitfury_info *info, int chip)
+{
+	int chips = chip + 1;
+	size_t old, new;
+
+	if (likely(chips <= info->chips))
+		return;
+	if (chips > 999)
+		return;
+	old = sizeof(int) * info->chips;
+	new = sizeof(int) * chips;
+	applog(LOG_INFO, "%s %d: Adjust chip size to %d", bitfury->drv->name, bitfury->device_id,
+	       chips);
+
+	recalloc(info->filtered_hw, old, new);
+	recalloc(info->job, old, new);
+	recalloc(info->submits, old, new);
+	if (info->chips == 2 && chips <= 6 && info->ident == IDENT_BXF)
+		bitfury->drv->name = "HXF";
+	else if (info->chips <= 6 && chips > 6 && info->ident == IDENT_BXF)
+		bitfury->drv->name = "MXF";
+	info->chips = chips;
+}
+
 static void parse_bxf_submit(struct cgpu_info *bitfury, struct bitfury_info *info, char *buf)
 {
 	struct work *match_work, *tmp, *work = NULL;
@@ -497,7 +901,11 @@ static void parse_bxf_submit(struct cgpu_info *bitfury, struct bitfury_info *inf
 		       bitfury->drv->name, bitfury->device_id);
 		return;
 	}
-	if (chip > -1 && chip < 2)
+	adjust_bxf_chips(bitfury, info, chip);
+	if (unlikely(chip >= info->chips || chip < 0)) {
+		applog(LOG_INFO, "%s %d: Invalid submit chip number %d",
+		       bitfury->drv->name, bitfury->device_id, chip);
+	} else
 		info->submits[chip]++;
 
 	applog(LOG_DEBUG, "%s %d: Parsed nonce %u workid %d timestamp %u",
@@ -562,7 +970,7 @@ static void parse_bxf_temp(struct cgpu_info *bitfury, struct bitfury_info *info,
 	}
 
 	mutex_lock(&info->lock);
-	info->temperature = (double)decitemp / 10;
+	bitfury->temp = (double)decitemp / 10;
 	if (decitemp > info->max_decitemp) {
 		info->max_decitemp = decitemp;
 		applog(LOG_DEBUG, "%s %d: New max decitemp %d", bitfury->drv->name,
@@ -600,7 +1008,7 @@ static void parse_bxf_temp(struct cgpu_info *bitfury, struct bitfury_info *info,
 			goto out;
 		}
 		/* implies: decitemp < info->last_decitemp */
-		if (info->clocks >= BXF_CLOCK_DEFAULT)
+		if (info->clocks >= opt_bxf_bits)
 			goto out;
 		applog(LOG_DEBUG, "%s %d: Temp %d in target and falling, increasing clock",
 		       bitfury->drv->name, bitfury->device_id, decitemp);
@@ -608,7 +1016,7 @@ static void parse_bxf_temp(struct cgpu_info *bitfury, struct bitfury_info *info,
 		goto out;
 	}
 	/* implies: decitemp < info->temp_target - BXF_TEMP_HYSTERESIS */
-	if (info->clocks >= BXF_CLOCK_DEFAULT)
+	if (info->clocks >= opt_bxf_bits)
 		goto out;
 	applog(LOG_DEBUG, "%s %d: Temp %d below target, increasing clock",
 		bitfury->drv->name, bitfury->device_id, decitemp);
@@ -643,7 +1051,8 @@ static void parse_bxf_job(struct cgpu_info *bitfury, struct bitfury_info *info,
 		       bitfury->drv->name, bitfury->device_id);
 		return;
 	}
-	if (chip > 1) {
+	adjust_bxf_chips(bitfury, info, chip);
+	if (chip >= info->chips || chip < 0) {
 		applog(LOG_INFO, "%s %d: Invalid job chip number %d",
 		       bitfury->drv->name, bitfury->device_id, chip);
 		return;
@@ -660,7 +1069,8 @@ static void parse_bxf_hwerror(struct cgpu_info *bitfury, struct bitfury_info *in
 		       bitfury->drv->name, bitfury->device_id);
 		return;
 	}
-	if (chip > 1) {
+	adjust_bxf_chips(bitfury, info, chip);
+	if (chip >= info->chips || chip < 0) {
 		applog(LOG_INFO, "%s %d: Invalid hwerror chip number %d",
 		       bitfury->drv->name, bitfury->device_id, chip);
 		return;
@@ -696,7 +1106,7 @@ static void *bxf_get_results(void *userdata)
 	bxf_update_work(bitfury, info);
 
 	while (likely(!bitfury->shutdown)) {
-		char *msg;
+		char *msg, *strbuf;
 		int err;
 
 		if (unlikely(bitfury->usbinfo.nodev))
@@ -711,14 +1121,25 @@ static void *bxf_get_results(void *userdata)
 		if (!err)
 			continue;
 
-		PARSE_BXF_MSG(submit);
+		if (opt_bxf_debug) {
+			strbuf = str_text(buf);
+			applog(LOG_ERR, "%s %d: < [%s]",
+				bitfury->drv->name, bitfury->device_id, strbuf);
+			free(strbuf);
+		}
+
+                PARSE_BXF_MSG(submit);
 		PARSE_BXF_MSG(temp);
 		PARSE_BXF_MSG(needwork);
 		PARSE_BXF_MSG(job);
 		PARSE_BXF_MSG(hwerror);
 
-		applog(LOG_DEBUG, "%s %d: Unrecognised string %s",
-		       bitfury->drv->name, bitfury->device_id, buf);
+		if (buf[0] != '#') {
+			strbuf = str_text(buf);
+			applog(LOG_DEBUG, "%s %d: Unrecognised string %s",
+			       bitfury->drv->name, bitfury->device_id, strbuf);
+			free(strbuf);
+		}
 	}
 out:
 	return NULL;
@@ -726,10 +1147,14 @@ out:
 
 static bool bxf_prepare(struct cgpu_info *bitfury, struct bitfury_info *info)
 {
+	bxf_send_ledmode(bitfury);
+	bxf_send_debugmode(bitfury);
+
 	mutex_init(&info->lock);
 	if (pthread_create(&info->read_thr, NULL, bxf_get_results, (void *)bitfury))
 		quit(1, "Failed to create bxf read_thr");
-	return bxf_send_clock(bitfury, info, BXF_CLOCK_DEFAULT);
+
+	return bxf_send_clock(bitfury, info, opt_bxf_bits);
 }
 
 static bool bitfury_prepare(struct thr_info *thr)
@@ -741,6 +1166,7 @@ static bool bitfury_prepare(struct thr_info *thr)
 
 	switch(info->ident) {
 		case IDENT_BXF:
+		case IDENT_OSM:
 			return bxf_prepare(bitfury, info);
 			break;
 		case IDENT_BF1:
@@ -876,11 +1302,14 @@ out:
 
 static int64_t bxf_scan(struct cgpu_info *bitfury, struct bitfury_info *info)
 {
+	int ms, aged;
 	int64_t ret;
-	int aged;
 
 	bxf_update_work(bitfury, info);
-	cgsleep_ms(600);
+	ms = 1200 / info->chips;
+	if (ms < 100)
+		ms = 100;
+	cgsleep_ms(ms);
 
 	mutex_lock(&info->lock);
 	ret = bitfury_rate(info);
@@ -902,28 +1331,42 @@ static int64_t bxf_scan(struct cgpu_info *bitfury, struct bitfury_info *info)
 	return ret;
 }
 
-static int64_t nf1_scan(struct thr_info *thr, struct cgpu_info *bitfury,
+static void bitfury_check_work(struct thr_info *thr, struct cgpu_info *bitfury,
+			       struct bitfury_info *info, int chip_n)
+{
+	if (!info->work[chip_n]) {
+		info->work[chip_n] = get_work(thr, thr->id);
+		if (unlikely(thr->work_restart)) {
+			free_work(info->work[chip_n]);
+			info->work[chip_n] = NULL;
+			return;
+		}
+		bitfury_work_to_payload(&info->payload[chip_n], info->work[chip_n]);
+	}
+
+	if (unlikely(bitfury->usbinfo.nodev))
+		return;
+
+	if (!libbitfury_sendHashData(thr, bitfury, info, chip_n))
+		usb_nodev(bitfury);
+
+	if (info->job_switched[chip_n]) {
+		if (likely(info->owork[chip_n]))
+			free_work(info->owork[chip_n]);
+		info->owork[chip_n] = info->work[chip_n];
+		info->work[chip_n] = NULL;
+	}
+
+}
+
+static int64_t nfu_scan(struct thr_info *thr, struct cgpu_info *bitfury,
 			struct bitfury_info *info)
 {
 	int64_t ret = 0;
+	int i;
 
-	if (!info->work) {
-		info->work = get_work(thr, thr->id);
-		if (unlikely(thr->work_restart)) {
-			free_work(info->work);
-			return 0;
-		}
-		bitfury_work_to_payload(&info->payload, info->work);
-	}
-	if (!libbitfury_sendHashData(thr, bitfury, info))
-		return -1;
-
-	if (info->job_switched) {
-		if (likely(info->owork))
-			free_work(info->owork);
-		info->owork = info->work;
-		info->work = NULL;
-	}
+	for (i = 0; i < info->chips; i++)
+		bitfury_check_work(thr, bitfury, info, i);
 
 	ret = bitfury_rate(info);
 
@@ -940,17 +1383,45 @@ static int64_t bitfury_scanwork(struct thr_info *thr)
 {
 	struct cgpu_info *bitfury = thr->cgpu;
 	struct bitfury_info *info = bitfury->device_data;
+	int64_t ret = -1;
+
+	if (unlikely(share_work_tdiff(bitfury) > 60)) {
+		if (info->failing) {
+			if (share_work_tdiff(bitfury) > 120) {
+				applog(LOG_ERR, "%s %d: Device failed to respond to restart",
+				       bitfury->drv->name, bitfury->device_id);
+				return ret;
+			}
+		} else {
+			applog(LOG_WARNING, "%s %d: No valid hashes for over 1 minute, attempting to reset",
+			       bitfury->drv->name, bitfury->device_id);
+			usb_reset(bitfury);
+			info->failing = true;
+		}
+	}
+
+	if (unlikely(bitfury->usbinfo.nodev))
+		return ret;
 
 	switch(info->ident) {
 		case IDENT_BF1:
-			return bf1_scan(thr, bitfury, info);
+			ret = bf1_scan(thr, bitfury, info);
+			break;
 		case IDENT_BXF:
-			return bxf_scan(bitfury, info);
-		case IDENT_NF1:
-			return nf1_scan(thr, bitfury, info);
+		case IDENT_OSM:
+			ret = bxf_scan(bitfury, info);
+			break;
+		case IDENT_NFU:
+		case IDENT_BXM:
+			ret = nfu_scan(thr, bitfury, info);
+			break;
 		default:
-			return 0;
+			ret = 0;
+			break;
 	}
+	if (ret > 0)
+		info->failing = false;
+	return ret;
 }
 
 static void bxf_send_maxroll(struct cgpu_info *bitfury, int maxroll)
@@ -998,6 +1469,7 @@ static void bitfury_flush_work(struct cgpu_info *bitfury)
 
 	switch(info->ident) {
 		case IDENT_BXF:
+		case IDENT_OSM:
 			bxf_send_flush(bitfury);
 			bxf_update_work(bitfury, info);
 			bxf_update_work(bitfury, info);
@@ -1013,6 +1485,7 @@ static void bitfury_update_work(struct cgpu_info *bitfury)
 
 	switch(info->ident) {
 		case IDENT_BXF:
+		case IDENT_OSM:
 			bxf_update_work(bitfury, info);
 		case IDENT_BF1:
 		default:
@@ -1038,11 +1511,12 @@ static struct api_data *bf1_api_stats(struct bitfury_info *info)
 	return root;
 }
 
-static struct api_data *bxf_api_stats(struct bitfury_info *info)
+static struct api_data *bxf_api_stats(struct cgpu_info *bitfury, struct bitfury_info *info)
 {
 	struct api_data *root = NULL;
 	double nonce_rate;
 	char buf[32];
+	int i;
 
 	sprintf(buf, "%d.%d", info->ver_major, info->ver_minor);
 	root = api_add_string(root, "Version", buf, true);
@@ -1051,19 +1525,35 @@ static struct api_data *bxf_api_stats(struct bitfury_info *info)
 	nonce_rate = (double)info->total_nonces / (double)info->cycles;
 	root = api_add_double(root, "NonceRate", &nonce_rate, true);
 	root = api_add_int(root, "NoMatchingWork", &info->no_matching_work, false);
-	root = api_add_double(root, "Temperature", &info->temperature, false);
+	root = api_add_double(root, "Temperature", &bitfury->temp, false);
 	root = api_add_int(root, "Max DeciTemp", &info->max_decitemp, false);
 	root = api_add_uint8(root, "Clock", &info->clocks, false);
-	root = api_add_int(root, "Core0 hwerror", &info->filtered_hw[0], false);
-	root = api_add_int(root, "Core1 hwerror", &info->filtered_hw[1], false);
-	root = api_add_int(root, "Core0 jobs", &info->job[0], false);
-	root = api_add_int(root, "Core1 jobs", &info->job[1], false);
-	root = api_add_int(root, "Core0 submits", &info->submits[0], false);
-	root = api_add_int(root, "Core1 submits", &info->submits[1], false);
+	for (i = 0; i < info->chips; i++) {
+		sprintf(buf, "Core%d hwerror", i);
+		root = api_add_int(root, buf, &info->filtered_hw[i], false);
+		sprintf(buf, "Core%d jobs", i);
+		root = api_add_int(root, buf, &info->job[i], false);
+		sprintf(buf, "Core%d submits", i);
+		root = api_add_int(root, buf, &info->submits[i], false);
+	}
 
 	return root;
 }
 
+static struct api_data *nfu_api_stats(struct bitfury_info *info)
+{
+	struct api_data *root = NULL;
+	char buf[32];
+	int i;
+
+	root = api_add_int(root, "Chips", &info->chips, false);
+	for (i = 0; i < info->chips; i++) {
+		sprintf(buf, "Core%d submits", i);
+		root = api_add_int(root, buf, &info->submits[i], false);
+	}
+	return root;
+}
+
 static struct api_data *bitfury_api_stats(struct cgpu_info *cgpu)
 {
 	struct bitfury_info *info = cgpu->device_data;
@@ -1073,7 +1563,12 @@ static struct api_data *bitfury_api_stats(struct cgpu_info *cgpu)
 			return bf1_api_stats(info);
 			break;
 		case IDENT_BXF:
-			return bxf_api_stats(info);
+		case IDENT_OSM:
+			return bxf_api_stats(cgpu, info);
+			break;
+		case IDENT_NFU:
+		case IDENT_BXM:
+			return nfu_api_stats(info);
 			break;
 		default:
 			break;
@@ -1087,11 +1582,10 @@ static void bitfury_get_statline_before(char *buf, size_t bufsiz, struct cgpu_in
 
 	switch(info->ident) {
 		case IDENT_BXF:
-			tailsprintf(buf, bufsiz, "%5.1fC         | ", info->temperature);
+		case IDENT_OSM:
+			tailsprintf(buf, bufsiz, "%5.1fC", cgpu->temp);
 			break;
-		case IDENT_BF1:
 		default:
-			tailsprintf(buf, bufsiz, "               | ");
 			break;
 	}
 }
@@ -1111,7 +1605,6 @@ static void bitfury_init(struct cgpu_info *bitfury)
 		case IDENT_BF1:
 			bf1_init(bitfury);
 			break;
-		case IDENT_BXF:
 		default:
 			break;
 	}
@@ -1133,14 +1626,19 @@ static void bitfury_shutdown(struct thr_info *thr)
 			bf1_close(bitfury);
 			break;
 		case IDENT_BXF:
+		case IDENT_OSM:
 			bxf_close(info);
 			break;
-		case IDENT_NF1:
-			nf1_close(bitfury);
+		case IDENT_NFU:
+			nfu_close(bitfury);
+			break;
+		case IDENT_BXM:
+			bxm_close(bitfury, info);
 			break;
 		default:
 			break;
 	}
+	usb_nodev(bitfury);
 }
 
 /* Currently hardcoded to BF1 devices */
