@@ -534,6 +534,8 @@ static void icarus_initialise(struct cgpu_info *icarus, int baud)
 			_transfer(icarus, CP210X_TYPE_OUT, CP210X_REQUEST_BAUD, 0,
 				 interface, &data, sizeof(data), C_SETBAUD);
 			break;
+ 		case IDENT_AVA:
+ 			break;
 		default:
 			quit(1, "icarus_intialise() called with invalid %s cgid %i ident=%d",
 				icarus->drv->name, icarus->cgminer_id, ident);
@@ -649,6 +651,7 @@ static void set_timing_mode(int this_option_offset, struct cgpu_info *icarus)
 	ident = usb_ident(icarus);
 	switch (ident) {
 		case IDENT_ICA:
+		case IDENT_AVA:
 			info->Hs = ICARUS_REV3_HASH_TIME;
 			read_count_timing = ICARUS_READ_COUNT_TIMING;
 			break;
@@ -828,6 +831,7 @@ static void get_options(int this_option_offset, struct cgpu_info *icarus, int *b
 		case IDENT_ICA:
 		case IDENT_BLT:
 		case IDENT_LLT:
+		case IDENT_AVA:
 			*baud = ICARUS_IO_SPEED;
 			*work_division = 2;
 			*fpga_count = 2;
@@ -1114,6 +1118,7 @@ static struct cgpu_info *icarus_detect_one(struct libusb_device *dev, struct usb
 	info->ident = usb_ident(icarus);
 	switch (info->ident) {
 		case IDENT_ICA:
+        case IDENT_AVA:
 		case IDENT_BLT:
 		case IDENT_LLT:
 		case IDENT_AMU:
@@ -1163,7 +1168,7 @@ cmr2_retry:
 			continue;
 
 		memset(nonce_bin, 0, sizeof(nonce_bin));
-		ret = icarus_get_nonce(icarus, nonce_bin, &tv_start, &tv_finish, NULL, 100);
+		ret = icarus_get_nonce(icarus, nonce_bin, &tv_start, &tv_finish, NULL, 300);
 		if (ret != ICA_NONCE_OK)
 			continue;
 
