@@ -85,6 +85,12 @@
 #define CP210X_VALUE_DATA 0x0303
 #define CP210X_DATA_BAUD 0x0001c200
 
+#define CP210X_SET_LINE_CTL 0x03
+#define CP210X_BITS_DATA_MASK 0x0f00
+#define CP210X_BITS_DATA_8 0x0800
+#define CP210X_BITS_PARITY_MARK 0x0030
+#define CP210X_BITS_PARITY_SPACE 0x0040
+
 
 // For 0x067b:0x2303 Prolific PL2303 - ICA
 #define PL2303_CTRL_DTR 0x01
@@ -139,14 +145,17 @@ enum sub_ident {
 	IDENT_ANU,
 	IDENT_AS2,
 	IDENT_AVA,
+	IDENT_AV2,
 	IDENT_BAJ,
 	IDENT_BAL,
 	IDENT_BAM,
 	IDENT_BAS,
 	IDENT_BBF,
+	IDENT_BET,
 	IDENT_BF1,
 	IDENT_BFL,
 	IDENT_BLT,
+	IDENT_BMA,
 	IDENT_BTB,
 	IDENT_BXF,
 	IDENT_BXM,
@@ -155,6 +164,7 @@ enum sub_ident {
 	IDENT_CTA,
 	IDENT_DRB,
 	IDENT_HFA,
+	IDENT_HRO,
 	IDENT_ICA,
 	IDENT_KLN,
 	IDENT_LIN,
@@ -289,8 +299,11 @@ struct cg_usb_info {
 	USB_ADD_COMMAND(C_PROGRAMSTATUS, "ProgramStatus") \
 	USB_ADD_COMMAND(C_PROGRAMSTATUS2, "ProgramStatus2") \
 	USB_ADD_COMMAND(C_FINALPROGRAMSTATUS, "FinalProgramStatus") \
-	USB_ADD_COMMAND(C_SETCLOCK, "SetClock") \
+	USB_ADD_COMMAND(C_SETCLOCK, "SetClock")			    \
+	USB_ADD_COMMAND(C_SETPARITY, "SetParity") \
 	USB_ADD_COMMAND(C_REPLYSETCLOCK, "ReplySetClock") \
+	USB_ADD_COMMAND(C_SETVOLT, "SetVolt") \
+	USB_ADD_COMMAND(C_REPLYSETVOLT, "ReplySetVolt") \
 	USB_ADD_COMMAND(C_REQUESTUSERCODE, "RequestUserCode") \
 	USB_ADD_COMMAND(C_GETUSERCODE, "GetUserCode") \
 	USB_ADD_COMMAND(C_REQUESTTEMPERATURE, "RequestTemperature") \
@@ -363,6 +376,10 @@ struct cg_usb_info {
 	USB_ADD_COMMAND(C_ATMEL_OPEN, "AtmelOpen") \
 	USB_ADD_COMMAND(C_ATMEL_INIT, "AtmelInit") \
 	USB_ADD_COMMAND(C_ATMEL_CLOSE, "AtmelClose") \
+	USB_ADD_COMMAND(C_AVA2_READ, "Ava2Read") \
+	USB_ADD_COMMAND(C_AVA2_WRITE, "Ava2Write") \
+        USB_ADD_COMMAND(C_BET_WRITE, "BlockErupterWrite") \
+        USB_ADD_COMMAND(C_BET_READ, "BlockErupterRead") \
 	USB_ADD_COMMAND(C_BF1_REQINFO, "BF1RequestInfo") \
 	USB_ADD_COMMAND(C_BF1_GETINFO, "BF1GetInfo") \
 	USB_ADD_COMMAND(C_BF1_REQRESET, "BF1RequestReset") \
@@ -419,6 +436,8 @@ struct cg_usb_info {
 	USB_ADD_COMMAND(C_HF_NOTICE, "HFNotice") \
 	USB_ADD_COMMAND(C_HF_PING, "HFPing") \
 	USB_ADD_COMMAND(C_HF_FAN, "HFFan") \
+	USB_ADD_COMMAND(C_HRO_WRITE, "HROWrite") \
+	USB_ADD_COMMAND(C_HRO_READ, "HRORead") \
 	USB_ADD_COMMAND(C_OP_NAME, "HFName") \
 	USB_ADD_COMMAND(C_HF_GETHEADER, "HFGetHeader") \
 	USB_ADD_COMMAND(C_HF_GETDATA, "HFGetData") \
