@@ -14,6 +14,7 @@
 	#define INVSOCK -1
 	#define INVINETADDR -1
 	#define CLOSESOCKET close
+	#define INET_PTON inet_pton
 
 	#define SOCKERRMSG strerror(errno)
 	static inline bool sock_blocks(void)
@@ -38,6 +39,9 @@
 	#define INVINETADDR INADDR_NONE
 	#define CLOSESOCKET closesocket
 
+	int Inet_Pton(int af, const char *src, void *dst);
+	#define INET_PTON Inet_Pton
+
 	extern char *WSAErrorMsg(void);
 	#define SOCKERRMSG WSAErrorMsg()
 
@@ -120,6 +124,8 @@ void us_to_timeval(struct timeval *val, int64_t us);
 void us_to_timespec(struct timespec *spec, int64_t us);
 void ms_to_timespec(struct timespec *spec, int64_t ms);
 void timeraddspec(struct timespec *a, const struct timespec *b);
+char *Strcasestr(char *haystack, const char *needle);
+char *Strsep(char **stringp, const char *delim);
 void cgsleep_ms(int ms);
 void cgsleep_us(int64_t us);
 void cgtimer_time(cgtimer_t *ts_start);
